;;; arithmatic.lisp -- basic belief function arithimatic functions

;;; Copyright 1986 Russell G. Almond
;;; License is granted to copy this program for education or research
;;; purposes, with the restriction that no portion of this program may
;;; be copied without also copying this notice.  All other rights
;;; reserved. 

;;; 12/15/86 -- changed to refect new ps-set functions
;;; 7/13/88 -- changed to use explicit **frame** value for frame
;;; 2/15/89 -- changed to allow alternative potential value and long
;;; precision arithmatic
;;; 8/5/89 -- Version 1.1
;;;	* Maximizing version of projection[d]
;;;	  new keyword (:op #'+) added to @v & @-> (& @^) defaults to
;;;	  #'+ but can be set to #'max
;;;	* Mixture operator[d]
;;; 8/4/89 -- added copy-change-frame which copies and changes frames,
;;; works on both belief functions and potentials
;(provide 'arithmatic)
(in-package 'belief :nicknames '(bel))
(require 'structures (merge-pathnames "structures" user::bel-home))
(require 'sets (merge-pathnames "sets" user::bel-home))
(use-package 'basic)
(use-package 'sets)
(export '(@^ @v @-> @+ @+2 || -><- ppval make-vacuous
	     direct-sum conflict-bel-fun normalize-bel-fun
	     approx-zerop ppbel node-size reset-zero-tol mix
	     copy-change-frame))
;; located in rules package, but initialized here are
;;  *min-input* *condition-ratio* *att-penalty* *zero-tol*

;;;; minimal extension

;;; @^ -- minimal extension -- extend bel-fun so that its frame is
;;; new-frame.  (note: this also permutes the orders of the variables.)
(defun @^ (new-frame bel-fun &key (op #'+))
  (declare (list new-frame) (type belief-function bel-fun))
  "(bel:@^ new-frame bel-fun) extends bel-fun to cover new-frame.
   New-frame must contain the frame of the belief function."
  (cond ((not (subsetp (belief-function-frame bel-fun) new-frame))
	 (error "bel:@^: illeagal frame ~S.  Try using bel:@->." new-frame))
	((equalp new-frame (belief-function-frame bel-fun)) bel-fun)
	(t (make-belief-function
	    :frame new-frame
	    :ms (extend-ms-list new-frame (belief-function-frame bel-fun)
				(belief-function-ms bel-fun))))))

;;; extend-ms-list -- minimally extends the focal elements in an
;;; ms-list to cover the new-frame.  It also permutes the order of the
;;; elements. Focal elements in the ms-list cover variables given in
;;; the old-frame argument.
(defun extend-ms-list (new-frame old-frame ms-list)
  (declare (list new-frame) (list old-frame) (list ms-list))
  (cond ((endp ms-list) nil)
	(t (cons (make-m-value
		  :m (m-value-m (car ms-list))
		  :element (extend-ps-sym new-frame old-frame
					  (m-value-element (car ms-list))))
		 (extend-ms-list new-frame old-frame (cdr ms-list))))))

;;; extend-ps-sym -- extends a single focal element to cover new frame
;;; instead of old frame.  The focal element is assument to be a symbol
;;; rather than an actual ps-set
(defun extend-ps-sym (new-frame old-frame ps-sym)
  (declare (list new-frame) (list old-frame) (symbol ps-sym))
  (if (eq ps-sym **frame**) **frame**
    (get-ps-set-sym (mapcar #'(lambda (ps-el)
				(copy-ps-el new-frame old-frame ps-el)) 
			    (ps-set-sym-val ps-sym)))))



;;; copy-ps-el -- copies one possible outcome.  As the outcome is
;;; copied, it is moved from the old frame to new frame.
(defun copy-ps-el (new-frame old-frame ps-el)
  (declare (list new-frame) (list old-frame) (list ps-el))
  (if (null ps-el) nil
    (do ((new-el (make-list  (length new-frame)))
	 (i 0 (1+ i)))
	((eql i (length new-frame)) new-el)
	(declare (fixnum i))
	(let ((offset (position (elt new-frame i) old-frame)))
	  (setf (elt new-el i) (if offset (elt ps-el offset)
				 (get (elt new-frame i) :values)
				 ;; assume variable have :values property
				 ))))))

;;;; Marginalization

;;; @v -- marginalization.  Finds the margin of bel in the new frame.
;;; This routine is optimized for the case when new-frame is a subset
;;; of the frame of the belief function.
(defun @v (new-frame bel-fun &key (op #'+))
  (declare (list new-frame) (type belief-function bel-fun))
  "(bel:@v new-frame bel-fun) Margin of bel-fun on new-frame.
   Assumes that new-frame is a subset of frame of bel-fun."
  (cond ((not (subsetp new-frame (belief-function-frame bel-fun)))
	 (error "bel:@v : Illegal new-frame ~S.  Use bel:@->"))
	(t (make-belief-function
	    :frame new-frame
	    :ms (remove-universe
		 new-frame
		 (colapse-ms-list
		  (margin-ms-list new-frame (belief-function-frame bel-fun)
				  (belief-function-ms bel-fun))
		  :op op))))))


;;; colapse-ms-list -- sums over duplicate entries in a list of
;;; m-values 
;; added op keyword which defines the reduction operation.  Currently
;; other interesting choice is #'max
(defun colapse-ms-list (ms-list &key (op #'+))
  (declare (list ms-list))
  (if (endp ms-list) nil
    (if (and (approx-zerop (m-value-m (car ms-list)))
	     (not (eq (m-value-element (car ms-list)) **frame**)))
	(colapse-ms-list (cdr ms-list))
      (cons (make-m-value
	     :element (m-value-element (car ms-list))
	     :m (reduce op (mapcar #'m-value-m
				   (remove (m-value-element (car ms-list))
					   ms-list
					   :key #'m-value-element
					   :test-not #'eq))))
	    (colapse-ms-list (remove (m-value-element (car ms-list))
				     ms-list
				     :key #'m-value-element
				     :test #'eq)
			     :op op)))))



;;; margin-ms-list -- find the margins of each of the element of the
;;; ms-list in the new-frame
(defun margin-ms-list (new-frame old-frame ms-list)
  (declare (list new-frame) (list old-frame) (list ms-list))
  (cond ((endp ms-list) nil)
	(t (cons (make-m-value
		  :m (m-value-m (car ms-list))
		  :element (margin-ps-sym new-frame old-frame
					  (m-value-element
					   (car ms-list))))
		 (margin-ms-list new-frame old-frame (cdr ms-list))))))


;;; margin-ps-sym -- creates a new ps-sym that covers the
;;; marginalization of the ps-set to the new frame
(defun margin-ps-sym (new-frame old-frame ps-sym)
  (declare (list new-frame) (list old-frame) (symbol ps-sym))
  (if (eq ps-sym **frame**) **frame**
    (fget-ps-set-sym
     (remove-duplicates
      (mapcar #'(lambda (ps-el) (copy-ps-el new-frame old-frame ps-el))
	      (ps-set-sym-val ps-sym))
      :test #'equalp))))



;;;; projection

;;; @-> -- projection  -- this projects bel-fun onto a new frame.  It
;;; is like (bel:@v new-frame
;;;                 (bel:@^ (union new-frame (b-f-frame bel-fun))
;;;                         bel-fun)) 
;;; but more efficient.
(defun @-> (new-frame bel-fun &key (op #'+))
  (declare (list new-frame) (type belief-function bel-fun))
  "(bel:@-> new-frame bel-fun) coerces bel-fun to be over the new
   frame by minimal extension and marginalization."
  (make-belief-function
   :frame new-frame
   :ms (remove-universe
	new-frame
	(colapse-ms-list
	 (extend-ms-list new-frame (belief-function-frame bel-fun)
			 (belief-function-ms bel-fun))
	 :op op))))




	
;;;; Direct sums

;;; direct-sum -- takes the direct sum of two belief functions over
;;; different frames.  The resulting frame is the union of the two
;;; frames. 
(defun direct-sum (bel1 bel2)
  (declare (type belief-function bel1) (type belief-function bel2))
  "Direct sum over different frames."
  (let ((new-nodes (union (belief-function-frame bel1)
			      (belief-function-frame bel2))))
    (@+ (@^ new-nodes bel1) (@^ new-nodes bel2))))

;;; direct-sum-operator -- takes the direct sum of two belief
;;; functions over the same frame.
(defun @+ (bel1 bel2 &key (op #'+) )
  (declare (type belief-function bel1) (type belief-function bel2)
	   (function op (number number) number))
  "Direct sum over the same frame."
  (if (not (equalp (belief-function-frame bel1)
		   (belief-function-frame bel2)))
      (error "bel:@+ : Belief functions do not have same frame.")
    (make-belief-function
     :frame (belief-function-frame bel1)
     :ms (rescale-ms-list
	  (colapse-ms-list
	   (ms-list-prod (belief-function-ms bel1) (belief-function-ms bel2))
	   :op op))
     )))

;;; direct-sum-operator-unnormalized -- takes the direct sum of two belief
;;; functions over the same frame.
(defun @+2 (bel1 bel2 &key (op #'+))
  (declare (type belief-function bel1) (type belief-function bel2)
	   (function op (number number) number))
  "Direct sum over the same frame."
  (if (not (equalp (belief-function-frame bel1)
		   (belief-function-frame bel2)))
      (error "bel:@+ : Belief functions do not have same frame.")
    (make-belief-function
     :frame (belief-function-frame bel1)
     :ms (colapse-ms-list
	  (ms-list-prod (belief-function-ms bel1) (belief-function-ms bel2))
	  :op op))
     ))


;;; normalize-bel-fun -- this normalizes a belief function
(defun normalize-bel-fun (bel1)
  (declare (type belief-function bel1))
  (make-belief-function
   :frame (belief-function-frame bel1)
   :ms (rescale-ms-list
	(belief-function-ms bel1))))

;;; conflict-bel-fun -- returns the ammount of conflict associated
;;; with an improper  belief function.
(defun conflict-bel-fun (bel)
  (declare (type belief-function bel))
  (let ((null-value
	 (find 'mt-set (belief-function-ms bel) :key #'m-value-element)))
    (if (null null-value) 0.0 (m-value-m null-value))))



;;; ms-list-prod -- takes the cross product of two ms-lists.  It
;;; assumes that there is an implicit universe m-value at the end of
;;; the list with weight 1.0 - sum of the other weights.
;;; 7/13/88 -- changed to explicit **frame** value.
(defun ms-list-prod (ms1 ms2)
  (declare (list ms1) (list ms2))
  (mapcan #'(lambda (mv) (ms-prod-set (m-value-m mv)
				      (m-value-element mv)
				      ms2))
	  ms1))
  

;;; ms-prod-set -- takes the product of a single element m-value with
;;; m mm and element mel with all of the m-values on ms-list.  
(defun ms-prod-set (mm mel ms-list2)
  (declare (long-float mm) (symbol mel) (list ms-list2))
  (cond ((approx-zerop mm) nil)
	((eq mel **frame**) (ms-prod-universe mm ms-list2))
	(t
	 (mapcar #'(lambda (mv)
		     (make-m-value
		      :m (the long-float (* mm (m-value-m mv)))
		      :element (ps-sym-intersect mel (m-value-element mv))))
		 ms-list2))))




;;; ms-prod-universe -- takes the product of the last element (the
;;; universe) with the list of ms-values
(defun ms-prod-universe (mm ms-list)
  (declare (long-float mm) (list ms-list))
  (if (approx-zerop mm) nil
    (mapcar #'(lambda (mv)
		(make-m-value
		 :m (the long-float (* mm (m-value-m mv)))
		 :element (m-value-element mv)))
	    ms-list)))
	       


;;; rescale-ms-list -- This divides all elements of the ms-list by the
;;; scaling factor and remove the mt-set elements.  The scaling factor
;;; is 1 over 1- the sum of the m-value of the mt-set factor element.
(defun rescale-ms-list (ms-list)
  (declare (list ms-list))
  (remove 0
	  (let ((mm (/ (reduce #'- 
			       (mapcar #'m-value-m
				       (remove 'mt-set ms-list
					       :test-not #'eq
					       :key #'m-value-element))
			       :initial-value 1.0))))
	    (declare (long-float mm))
	    (mapcar #'(lambda (mv)
			(make-m-value
			 :m (* mm (m-value-m mv))
			 :element (m-value-element mv)))
		    (remove 'mt-set ms-list
			    :key #'m-value-element
			    :test #'eq)))
	  :key #'m-value-m
	  :test #'eql))


;;; remove-universe -- removes the universe by testing for length eql
;;; to product of the lengths of the :values props of the frame
(defun remove-universe (frame ms-list)
  (declare (list frame) (list ms-list))
  (let* ((uni-length (reduce #'* (mapcar #'(lambda (#1=#:x)
					     (length (the list (get #1# :values))))
			      frame)))
	 (frame-weight
	  (reduce #'+
		  (mapcar #'m-value-m
			  (remove-if-not #'(lambda (mv)
					     (is-frame mv uni-length)) ms-list ))
		  :initial-value 0.0))
	 (trim-msl 
	  (remove-if #'(lambda (mv)
			 (is-frame mv uni-length))  ms-list)))

     (append trim-msl (list (make-m-value :m frame-weight :element **frame**)))))

    




;;; is-frame -- used to test to see if a m-value-element represents
;;; the whole frame
(defun is-frame (#2=#:y frame-length)
  (declare (type m-value #2#) (number frame-length))
  (cond ((eq **frame** (m-value-element #2#)))
	((eql (length (ps-set-sym-val (m-value-element #2#))) frame-length))
	(t nil)))


;;; approx-zerop -- decides if the value is approximately equal to
;;; zero and if so returns t.
(defun approx-zerop (#1=#:x)
  (declare (long-float #1#))
  (if (< (abs #1#) (the float #!*zero-tol*))
      t nil))				;Hi tippit!


;;; *zero-tol* -- number which anything smaller than this should be
;;; taken as zero.  
(defvar *zero-tol* long-float-epsilon)

;;; *min-input* -- This variable is the minimum m-value input in any
;;; belief function
(defvar *min-input* .01)

;;; *condition-ratio* -- The condition ratio divided by the min-input
;;; is the smallest value that will be considered non-zero
(defvar *condition-ratio* 1.0E6)


(defun reset-zero-tol ()
  (set #?*zero-tol* (max (/ #!*min-input* #!*condition-ratio*)
			long-float-epsilon)))


;;; Things moved over from computations because they really belong in
;;; package belief

;;; ppbel -- prints a belief function in a human readable form
(defun ppbel (bel-fun &optional (stream t))
  (declare (type belief-function bel-fun))
  (format stream "~%Belief function over frame:~S~%"
	  (belief-function-frame bel-fun))
  (format stream "~&m-value~20Tfocal element~%")
  (map nil
   #'(lambda (mv)
       (format stream "~&~16,9G~20T{ " (m-value-m mv))
       (if (eq (m-value-element mv) **frame**) (format t "**frame** }~%")
	 (format stream "~{~#[~;~S ~; ~S ~S~; ~S ~S ~S ~; ~S ~S ~S ~S ~:;~S ~S ~S ~S~%~20T~]~} }~%"
		 (ps-set-sym-val (m-value-element mv)))))
   (belief-function-ms bel-fun))
  (terpri stream))

(defun ppval (bel-fun &optional (stream t)) (ppbel bel-fun stream))
(defun || (bel-fun) (normalize-bel-fun bel-fun))
(defun -><- (bel-fun) (conflict-bel-fun bel-fun))

;;; node-size -- This caluclates the size of a node, where a node is a
;;; list of attributes
(defun node-size (node)
  (declare (list node))
  (reduce #'*
	  (mapcar #'(lambda (att)
		      (+ (length (the list (get att :values))) #!*att-penalty*))
		  node)))

(defvar *att-penalty* 0 "Penalty for more attributes in a node")	  

(defun make-vacuous (frame)
  (make-belief-function :frame frame :ms vacuous))



;;; copy-change-frame -- copies a belief function structure and chages
;;; its name at the same time.
(defun copy-change-frame (frame bel-fun)
  (declare (list frame) (type belief-function bel-fun))
  (let ((res (copy-belief-function bel-fun)))
    (setf (belief-function-frame res) frame)
    res))


;;; mix -- forms a weighted mixture of several belief functions.
;;; Note: as this uses mapcan to do its basic stuff, if the two lists
;;; are of unequal length, the excess elements of the longer one will
;;; be ignored.
(defun mix (list-bels list-weights)
  (declare (list list-bels) (list list-weights))
  (when (find (val-frame (car list-bels))
	      (cdr list-bels) :key #'val-frame
	      :test-not #'equal)
	(error "mix:All values must have the same frame"))
  (make-belief-function :frame (val-frame (car list-bels))
       :ms (colapse-ms-list
	    (mapcan #'rescale-ms list-bels list-weights)
	    :op #'+)))

;; rescale-ms -- takes a belief function and a weight and returns a
;; new-ms-list formed by copying and rescaling all of the elements of
;; the old list.
(defun rescale-ms (bel-fun weight)
  (declare (type belief-function bel-fun) (number weight))
  (mapcar #'(lambda (ms-val &aux (nm-val (copy-m-value ms-val)))
	      (setf (m-value-m nm-val) (the long-float (* weight (m-value-m nm-val))))
	      nm-val)
	  (belief-function-ms bel-fun)))



;;; provide when loaded
(provide 'arithmatic)
