;;; potentials.lisp -- basic potential function arithimatic functions

;;; Copyright 1989 Russell G. Almond
;;; License is granted to copy this program for education or research
;;; purposes, with the restriction that no portion of this program may
;;; be copied without also copying this notice.  All other rights
;;; reserved. 

;;; 2/15/89 -- Created file from arithmatic.lisp
;;; 8/5/89 -- Version 1.1
;;;	* Maximizing version of projection[d]
;;;	  new keyword (:op #'+) added to @v & @-> (& @^) defaults to
;;;	  #'+ but can be set to #'max
;;;	* Mixture operator[d]
;;; 8/4/89 -- added copy-chage-frame utility to copy a val while
;;; changing its frame.
;(provide 'potentials)
(in-package 'prob)
(require 'structures (merge-pathnames "structures" user::bel-home))
(require 'sets (merge-pathnames "sets" user::bel-home))
(require 'lowread (merge-pathnames "lowread" user::bel-home))
(require 'utils (merge-pathnames "utils" user::bel-home))
(use-package 'basic)
(use-package 'sets)
(use-package 'utils)
(export '(@^ @v @-> @+ @+2 || -><- ppval make-vacuous
	     direct-sum constant-pot normalize-pot
	     approx-zerop pppot node-size reset-zero-tol
	     copy-change-frame mix))
;; the following have initial values defined here, the real constants
;; are avaliable only in the *rules-package* (through #? and #!)
;;  *min-input* *condition-ratio* *att-penalty* *zero-tol*

;;;; minimal extension

;;; @^ -- minimal extension -- extend pot so that its frame is
;;; new-frame.  (note: this also permutes the orders of the variables.)
(defun @^ (new-frame pot &key (op #'+))
  (declare (list new-frame) (type pot pot))
  "(prob:@^ new-frame pot) extends pot to cover new-frame.
   New-frame must contain the frame of the potential."
  (cond ((not (subsetp (pot-frame pot) new-frame))
	 (error "prob:@^: illeagal frame ~S.  Try using prob:@->." new-frame))
	((equalp new-frame (pot-frame pot)) pot)
	(t (make-pot
	    :frame new-frame
	    :array (make-array (frame-size (attribute-to-frame new-frame))
			       :element-type 'long-float
			       :initial-contents (array-constr-extend
						  new-frame 
						  (pot-frame pot)
						  pot))))))


;;; array-constr-extend -- creates and new array by extending and
;;; repeating the old arrays.  At each level of decent, if the new
;;; attribute is in the old frame, substitute each value in turn for
;;; the attribute in the old-frame-index.  At the base case,
;;; old-frame-index will then be an index into the array of potential,
;;; return that value.  If the attribute is not on the list, then just
;;; repeat one value for each new attribute. 
(defun array-constr-extend (undone-frame old-frame-index pot)
  (declare (list undone-frame) (list old-frame-index)
	   (type pot pot))
  (if (endp undone-frame)
      (elt-array (pot-array pot)
		 (car (ps-index (list old-frame-index)
			   (attribute-to-frame (pot-frame pot)))))
    (let* ((att (car undone-frame))
	   (ind (position att old-frame-index))
	   (new-frame-index (copy-seq old-frame-index)))
      (declare (list new-frame-index))
      (if ind
	  (mapcar #'(lambda (val)
		      (setf (elt new-frame-index ind) val)
		      (array-constr-extend (cdr undone-frame)
					   new-frame-index
					   pot))
		  (get att :values))
	(mapcar #'(lambda (val)
		    (array-constr-extend (cdr undone-frame)
					 new-frame-index
					 pot))
		(get att :values))))))


      



;;;; Marginalization

;;; @v -- marginalization.  Finds the margin of pot in the new frame.
;;; This routine is optimized for the case when new-frame is a subset
;;; of the frame of the potential.
(defun @v (new-frame pot &key (op #'+))
  (declare (list new-frame) (type pot pot))
  "(prob:@v new-frame pot) Margin of pot on new-frame.
   Assumes that new-frame is a subset of frame of pot."
  (cond ((not (subsetp new-frame (pot-frame pot)))
	 (error "prob:@v : Illegal new-frame ~S.  Use prob:@->"))
	(t (make-pot
	    :frame new-frame
	    :array (make-array (frame-size (attribute-to-frame new-frame))
			       :element-type 'long-float
			       :initial-contents (array-constr-marg
						  new-frame 
						  (pot-frame pot)
						  pot
						  :op op))))))



;;; array-constr-marg -- creates and new array by summing over
;;; the old arrays.  For each attribute of the old frame which is not
;;; in the new frame, substitute the list of possible values for the
;;; attribute.  Then pass the resulting list to array-constr-marg-aux
;;; which will calculate the result by recursive decent.
;;; At each level of decent, each value of the first attribute is
;;; substituted for the attribute name in the old-frame-index.  At the base case,
;;; old-frame-index will then be a a ps-set describing all the array
;;; elements which must be summed over to get the element for that
;;; postition in the new array.
(defun array-constr-marg (undone-frame old-frame-index pot &key (op #'+))
  (declare (list undone-frame) (list old-frame-index)
	   (type pot pot))
  (let ((new-frame-index (copy-seq old-frame-index)))
    (map nil #'(lambda (att)
		 (setq new-frame-index
		       (nsubstitute (get att :values) att new-frame-index)))
	  (set-difference old-frame-index undone-frame))
    (array-constr-marg-aux undone-frame new-frame-index pot :op op)))


(defun array-constr-marg-aux (undone-frame old-frame-index pot &key (op #'+))
  (declare (list undone-frame) (list old-frame-index)
	   (type pot pot))
  (if (endp undone-frame)
      (reduce op
	      (mapcar #'(lambda (index) (elt-array (pot-array pot) index))
		      (ps-index (list old-frame-index)
				(attribute-to-frame (pot-frame pot)))))
    (let* ((att (car undone-frame))
	   (ind (position att old-frame-index))
	   (new-frame-index (copy-seq old-frame-index)))
      (declare (list new-frame-index))
      (mapcar #'(lambda (val)
		  (setf (elt new-frame-index ind) val)
		  (array-constr-marg-aux (cdr undone-frame)
				       new-frame-index
				       pot :op op))
	      (get att :values)))))






;;;; projection

;;; @-> -- projection  -- this projects pot onto a new frame.  It
;;; is like (prob:@v new-frame
;;;                 (prob:@^ (union new-frame (b-f-frame pot))
;;;                         pot)) 
;;; but more efficient.
;;; added :op keyword so that maximizing projection (Pearls
;;; explanation stuff) could be used
(defun @-> (new-frame pot &key (op #'+))
  (declare (list new-frame) (type pot pot))
  "(prob:@-> new-frame pot) coerces pot to be over the new
   frame by minimal extension and marginalization."
  (make-pot
   :frame new-frame
   :array (make-array (frame-size (attribute-to-frame new-frame))
		      :element-type 'long-float
		      :initial-contents (array-constr-proj
					 new-frame 
					 (pot-frame pot)
					 pot :op op))))



;;; array-constr-proj -- creates and new array by summing over
;;; the old arrays.  For each attribute of the old frame which is not
;;; in the new frame, substitute the list of possible values for the
;;; attribute.  Then pass the resulting list to array-constr-proj-aux
;;; which will calculate the result by recursive decent.
;;; At each level of decent,  if the first attribute on the new
;;; attribute list is in the old-attribute list, then each value of
;;; the first attribute is substituted for the attribute name in the
;;; old-frame-index, otherwise the value is repeated for each
;;; instantiation of the new attribute.  At the base case, 
;;; old-frame-index will then be a a ps-set describing all the array
;;; elements which must be summed over to get the element for that
;;; postition in the new array.
(defun array-constr-proj (undone-frame old-frame-index pot
				       &key (op #'+))
  (declare (list undone-frame) (list old-frame-index)
	   (type pot pot))
  (let ((new-frame-index (copy-seq old-frame-index)))
    (map nil #'(lambda (att)
		 (setq new-frame-index
		       (nsubstitute (get att :values) att new-frame-index)))
	    (set-difference old-frame-index undone-frame))
    (array-constr-proj-aux undone-frame new-frame-index pot :op op)))



(defun array-constr-proj-aux (undone-frame old-frame-index pot
					   &key (op #'+))
  (declare (list undone-frame) (list old-frame-index)
	   (type pot pot))
  (if (endp undone-frame)
      (reduce op
	      (mapcar #'(lambda (index) (elt-array (pot-array pot) index))
		      (ps-index (list old-frame-index)
				(attribute-to-frame (pot-frame pot)))))
    (let* ((att (car undone-frame))
	   (ind (position att old-frame-index))
	   (new-frame-index (copy-seq old-frame-index)))
      (declare (list new-frame-index))
      (if ind 
	  (mapcar #'(lambda (val)
		      (setf (elt new-frame-index ind) val)
		      (array-constr-proj-aux (cdr undone-frame)
				       new-frame-index
				       pot :op op))
	      (get att :values))
	(mapcar #'(lambda (val)
		    (array-constr-proj-aux (cdr undone-frame)
					 new-frame-index
					 pot :op op))
		(get att :values))))))




	
;;;; Direct sums

;;; direct-sum -- takes the direct sum of two potentials over
;;; different frames.  The resulting frame is the union of the two
;;; frames. 
(defun direct-sum (prob1 prob2)
  (declare (type pot prob1) (type pot prob2))
  "Direct sum over different frames."
  (let ((new-nodes (union (pot-frame prob1)
			      (pot-frame prob2))))
    (@+ (@^ new-nodes prob1) (@^ new-nodes prob2))))

;;; direct-sum-operator-unnormalized -- takes the direct sum of two potentials
;;;  over the same frame, unnormalized
(defun @+2 (prob1 prob2 &key (op #'+))
  (declare (type pot prob1) (type pot prob2))
  "Direct sum over the same frame."
  (if (not (equalp (pot-frame prob1)
		   (pot-frame prob2)))
      (error "prob:@+ : potentials do not have same frame.")
    (make-pot
	 :frame (pot-frame prob1)
	 :array (prod-array (pot-array prob1) (pot-array prob2)))))


;; prod-array -- this function find the products of two arrays. 
(defun prod-array (arr1 arr2)
  (declare (array arr1) (array arr2))
  (as-array (array-dimensions arr1)
	    (coerce			;Damn Dec Anyway!  (map
					;'(vector long-float) returns
					;a '(vector t) in VAX lisp,
					;this kludge avoids that.
	     (map 'list #'* (as-vector arr1) (as-vector arr2))
	     '(vector long-float))
	    'long-float))




;;; direct-sum-operator-normalized -- takes the direct sum of two 
;;; potentials over the same frame.
(defun @+ (prob1 prob2  &key (op #'+))
  (declare (type pot prob1) (type pot prob2))
  "Direct sum over the same frame."
  (|| (@+2 prob1 prob2 :op op)))


;;; normalize-pot -- this normalizes a potential
(defun normalize-pot (prob1)
  (declare (type pot prob1))
  (make-pot
   :frame (pot-frame prob1)
   :array (rescale-array
	   (pot-array prob1))))

;;; rescale-array -- divides all elements of the array by the
;;; normalizing constant (sum of the elements)
(defun rescale-array (arr)
  (declare (array arr))
  (let* ((vec (as-vector arr))
	 (k (reduce #'+ vec)))
    (declare (float k))
    (when (eql k 0)
	(error "rescale-array: Divide by zero"))
    (as-array (array-dimensions arr)
	      (coerce			;Damn DEC Anyway!
	       (map 'list #'(lambda (#1=#:x) (/ #1# k)) vec)
	       '(vector long-float))
	      'long-float)))

	 

;;; constant-pot -- returns the normalizing constant associated
;;; with an improper  potential.
(defun constant-pot (prob)
  (declare (type pot prob))
  (reduce #'+ (as-vector (pot-array prob))))


;;; approx-zerop -- decides if the value is approximately equal to
;;; zero and if so returns t.
(defun approx-zerop (#1=#:x)
  (declare (long-float #1#))
  (if (< (abs #1#) (the float #!*zero-tol*))
      t nil))				;Hi tippit!

      

;;; *zero-tol* -- number which anything smaller than this should be
;;; taken as zero.  
(defvar *zero-tol* long-float-epsilon)

;;; *min-input* -- This variable is the minimum m-value input in any
;;; belief function
(defvar *min-input* .01)

;;; *condition-ratio* -- The condition ratio divided by the min-input
;;; is the smallest value that will be considered non-zero
(defvar *condition-ratio* 1.0E6)


(defun reset-zero-tol ()
  (set #?*zero-tol* (max (/ #!*min-input* #!*condition-ratio*)
			long-float-epsilon)))


;;; Things moved over from computations because they really belong in
;;; package prob

;;; pppot -- prints a potential in a human readable form
(defun pppot (pot &optional (stream t))
  (declare (type pot pot))
  (format stream "~%Potential over frame:~S~%"
	  (pot-frame pot))
  (pp-labeled-array (attribute-to-frame (pot-frame pot))
		    (pot-array pot) 0 #\( stream)
  (terpri stream))

;;; pp-labeled-array -- designed specificly as a recursive auxilary
;;; for printing potentials, but can be used to format/print arrays of
;;; number with labels for dimension.  labs is a list of labels for
;;; each dimension, its length must be equal to the rank of the array,
;;; and the length of each element must be equal to the corresponding
;;; dimension of the array.  tab gives the level of indentation the
;;; printing is currently at and pchar gives the opening parenthesis
;;; character which should be one of #\( or #\[.
(defun pp-labeled-array (labs arr &optional (tab 0) (pchar #\()
			      (stream t))
  (declare (list labs) (array arr) (fixnum tab) (character pchar)
	   (stream stream))
  (if (> tab 70)
      (setq tab 0 pchar (if (eql pchar #\() #\[ #\())
      ;line to long, wrap and change brackets
    )
  (if (endp labs) (return-from pp-labeled-array nil))
  (let* ((level-labs (car labs))
	 (rest-labs (cdr labs))
	 (fstring (if (endp rest-labs)
		      "~21,1,1<~*~:;~S~;~16,9G~>"
		    "~*~4,2@<~S~> "))
	 (maxlen (apply #'max
			(mapcar #'(lambda (item)
				    (length (the simple-string
						 (format nil " ~4,2@<~S~> " item))))
				level-labs)))
	 (sub-dims (unless (endp rest-labs)
			   (frame-size rest-labs)))
	 (sub-tot-size (reduce #'* sub-dims))
	 (printer (if (endp rest-labs)
		      #'(lambda (num)
			  (format stream fstring tab
				  (elt level-labs num)
				  (elt arr num))
			  (setq fstring "~22,1,<~%~vT~:; ~S~;~16,9G~>"))
		    #'(lambda (num)
			(format stream fstring tab (elt level-labs num))
			(setq fstring "~%~vT ~4,2@<~S~> ")
			(pp-labeled-array rest-labs
					  (make-array sub-dims
						      :element-type
						      (array-element-type arr)
						      :displaced-to arr
						      :displaced-index-offset
						      (* sub-tot-size num))
					  (the fixnum (+ tab maxlen)) pchar stream)))))
    (declare (list level-labs) (fixnum maxlen sub-tot-size)
	     (list sub-dims))
    (format stream "~C" pchar)
    (mapcar printer (iota (the fixnum (1- (length level-labs)))))
    (format stream "~C" (if (eql pchar #\() #\) #\]))))



(defun ppval (pot &optional (stream t)) (pppot pot stream))
(defun || (pot) (normalize-pot pot))
(defun -><- (pot) (constant-pot pot))

;;; node-size -- This caluclates the size of a node, where a node is a
;;; list of attributes
(defun node-size (node)
  (declare (list node))
  (reduce #'*
	  (mapcar #'(lambda (att) (+ (length (the list (get att :values))) #!*att-penalty*))
		  node)))

(defvar *att-penalty* 0 "Penalty for more attributes in a node")	  

(defun make-vacuous (frame)
  (make-pot :frame frame
	     :array (make-array (frame-size (attribute-to-frame frame))
				:initial-element 1.0L0
				:element-type 'long-float)))







;;; copy-change-frame -- copies a potential structure and chages
;;; its name at the same time.
(defun copy-change-frame (frame pott)
  (declare (list frame) (type pot pott))
  (let ((res (copy-pot pott)))
    (setf (pot-frame res) frame)
    res))



;;; mix -- forms a weighted mixture of several potentials
;;; Note: as this uses mapcan to do its basic stuff, if the two lists
;;; are of unequal length, the excess elements of the longer one will
;;; be ignored.
(defun mix (list-pots list-weights)
  (declare (list list-pots) (list list-weights))
  (when (find (val-frame (car list-pots))
	      (cdr list-pots) :key #'val-frame
	      :test-not #'equal)
	(error "mix:All values must have the same frame"))
  (make-pot :frame (val-frame (car list-pots))
	    :array (linear-comb (mapcar #'pot-array list-pots)
				list-weights)))



;;; provide when loaded
(provide 'potentials)