;;; utils --- genericly useful lisp utilities
;;;
;;; Copyright 1988 Russell G. Almond
;;; License is granted to copy this program for education or research
;;; purposes, with the restriction that no portion of this program may
;;; be copied without also copying this notice.  All other rights
;;; reserved. 


;;; 8/5/89 -- added linear combination operator used in prob package
;;; to form mixture potentials

;(provide 'utils)
(in-package 'utils )
(export '(glue-names add-count make-pairs over-common-cdrs ++
	  iota binomial ! add+  elt-array set-elt-array
	  as-vector as-array float-array long-float linear-comb
	  2+ 3+ 4+ 5+ 6+ 7+ 8+ 9+ 10+ 11+ 12+ 13+
	  fifo-list make-fifo-list fifo-list-p fifo-head fifo-last
	  fifo-pop fifo-append))


;; purge-list -- This function takes two arguments, a list and a
;; predicate.  It returns a list whos values consist of all those
;; members of the list with non null values of the predicate.
;(defun purge-list (test- list)
;  (declare (function test- (t) t) (list list))
;  (cond ((endp list) nil)
;	((eval (test- (car list)))
;	 (cons (car list) (purge-list test- (cdr list))))
;	(t (purge-list test- (cdr list)))))


;;remove-repeats -- This function takes a list and remove from it any
;;repeated (in the equal sense) elements.



;;; Name manipulation functions

;;; glue-names -- This function makes a new symbol with the names
;;; pasted together.
(defun glue-names (n1 n2)
  (declare (symbol n1) (symbol n2))
  (intern (concatenate 'string (symbol-name n1) "-" (symbol-name n2))))

;;; add-count -- This adds a reference count to a symbol name so that
;;; we are sure that we are getting unique names.  It adds a property
;;; :ref-count to the symbol.
(defun add-count (sym package)
  (declare (symbol sym))
  (cond ((null (get sym :ref-count))
	 (setf (get sym :ref-count) 0))
	((numberp (get sym :ref-count))
	 (incf (get sym :ref-count)))
	(t (error "add-count: symbol ~S, has refcount ~S" sym (get sym :ref-count))))
  (intern (concatenate 'string (symbol-name sym) "-" (write-to-string
						      (get sym :ref-count)))
	  package))




;;; make-pairs -- This takes a list with an even number of elements
;;; and pairs up the elements.
;;; 8/8/89 -- changed to return two values, the seond of which is the
;;; odd element out.
(defun make-pairs (list)
  (declare (list list))
  (cond ((endp list) '())
	((eql 1 (length list))
	 (values '() (car list)))
	(t (multiple-value-bind (p-list odd-el)
		     (make-pairs (cddr list))	
		     (values (cons (list (car list) (cadr list)) p-list)
			     odd-el)))))

	

;;; over-common-cdrs -- This takes an alist of which the car is
;;; considered to be some kind of value and the cdr is the units.  It
;;; applies the summary operation (op) to the values with the same
;;; units, and produces a reduceded alist.
(defun over-common-cdrs (op united-list)
  (declare (function op (&rest t) t) (list united-list))
  (if (endp united-list) nil
    (let ((unit (cdar united-list)))
      (cons (cons
	     (apply op
		    (remove unit united-list :key #'cdr :test-not #'equal))
	     unit)
	    (over-common-cdrs op
			      (remove unit united-list :key #'cdr
				      :test #'equal))))))


;;; ++ --- This function adds a list of two numbers together
(defun ++ (#1=#:x #2=#:y &rest more)
  (if more (apply #'++   (list (+ (car #1#) (car #2#))
			       (+ (cadr #1#) (cadr #2#)))
		  more)
    (list (+ (car #1#) (car #2#))
	  (+ (cadr #1#) (cadr #2#)))))


;; iota -- makes a list from i to j (if add+ is true, makes list from
;; i to j+)
(defun iota (i &optional (j 0) (add+ nil))
  (declare (fixnum i) (fixnum j))
  (if (< i j) (error "iota: negative argument ~S" i))
  (if add+
      (iota-aux (the fixnum (1- i)) j (list (intern (concatenate 'string
						    (write-to-string i) "+")
				       :user)) )
    (iota-aux i j nil))
  )


(defun iota-aux (i j rest)
  (declare (fixnum i j) (list rest))
  (cond ((< i j) rest)
	((= i j) (cons i rest))
	(t (iota-aux (1- i) j (cons i rest)))))


;; add+ -- if  i eq lim return i+ else return i
(defun add+ (i lim)
  (declare (integer i) (integer lim))
  (if (eq i lim) (values (intern (concatenate 'string (write-to-string i) "+")))
    i))

;;; binomial (n,m) -- finds the value of (m chose n)
(defun binomial (n m)
  (declare (fixnum n) (fixnum m))
  (cond ((zerop m) 1)
	((eql m n) 1)
	((minusp m) 0)
	((> m n) 0)
	((< (the fixnum (- n m)) m) (binomial n (the fixnum (- n m))))
	(t (binomial-aux n m))))

(defun binomial-aux (n m)
  (declare (fixnum n) (fixnum m))
  (if (zerop m) 1
    (/ (* (the fixnum (1+ (the fixnum (- n m))))
	  (binomial-aux n (the fixnum (1- m)))) m)))


;;; ! -- factorial
(defun ! (n)
  (declare (fixnum n))
  (cond ((minusp n) 0)
	((zerop n) 1)
	(t (* n (! (the fixnum (1- n)))))))


;;; These functions deal with setting elements of an array using a
;;; list of the indexies rather than an including the index of the
;;; lists directly.
(defmacro elt-array (array indexlist)
  `(apply #'aref ,array ,indexlist))

(defmacro set-elt-array (array indexlist value)
  `(setf (apply #'aref ,array ,indexlist) ,value))

(defsetf elt-array set-elt-array )


;;; Type casting functions to switch between arrays and vectors.

(defun as-vector (arr &optional (type t))
  (declare (array arr))
  (make-array (array-total-size arr) :displaced-to arr
	      :element-type (array-element-type arr))) 
						       


(defun as-array (dims vec &optional (type t))
  (declare (vector vec) (list dims))
  (make-array dims :displaced-to vec
	      :element-type (array-element-type vec))) 

;;; linear-comb -- forms the linear combination of serval arrays given
;;; a series of co-efficients.  As this thing use mapcar to do its
;;; basic operation, if the lists are of unequal length the shorter
;;; one is used.  
(defun linear-comb (array-list weight-list)
  (declare (list array-list) (list weight-list))
  (let ((vec-list (mapcar #'as-vector array-list)))
    (as-array (array-dimensions (car array-list))
	      (coerce				;Damn DEC anyway!
	       (eval `(map 'list #'(lambda (&rest args)
				     (weight-sum ',weight-list args))
			   ,.vec-list))
	       '(vector long-float))
	      'long-float)))

;; weight-sum -- takes a list of weights and then the rest of the
;; arguments are assumed to be things to be summed with the weights
(defun weight-sum (weight-list obj-list)
  (reduce #'+ (mapcar #'* weight-list obj-list)))



;;; Type casting function to create arrays of a specific type.

(defun float-array (arr)
  (declare (array arr))
  (as-array (array-dimensions arr)
	    (coerce (as-vector arr) '(vector long-float))
	    'long-float))

(defmacro long-float (num)
  `(float ,num 1.0L0))



;;; fifo-lists  -- this is a list with a pointer to both the first and
;;; last cons cell, so that the list can be rapidly (destructively)
;;; appended to and yet still items can be take from it in order
;; fifo-list make-fifo-list fifo-list-p fifo-head fifo-tail
;; fifo-pop fifo-append

(defstruct (fifo-list (:conc-name fifo-)
		      (:constructor make-fifo-list
				    (head &aux (tail (last head)))))
  (head nil :type list)
  (tail nil :type list))		;note tail point to last cons
					;cell not last item


;; fifo-pop -- pops an item from off the list and updates the
;; pointers.
(defun fifo-pop (fifo-list)
  (declare (type fifo-list fifo-list))
  (cond ((pop (fifo-head fifo-list)))
	(t (setf (fifo-tail fifo-list) nil) nil)))

;; fifo-append -- appends a list to a fifo-list and returns fifo-list 
(defun fifo-append (fifo-lst lst)
  (declare (type fifo-list fifo-lst) (list lst))
  (if (fifo-tail fifo-lst)
      (setf (cdr (fifo-tail fifo-lst)) lst)
    (setf (fifo-head fifo-lst) lst))
  (setf (fifo-tail fifo-lst) (last lst))
  fifo-lst)
;;; provide when loaded
(provide 'utils)